%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File: CRAY-LAPOPT-2
%
%  Description: Cray Code Optimization
%               Reformation of jump sequences
%               Unrolling of loops and tail recursions
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        22 Dec 1986
%
% Copyright (c) 1987,  Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% operation:
%
% LapoptReformJumps modifies jump sequences of the type
%         JSZ   L0005           or another conditional jump
%         J     anywhere
%   L0005                       target label follows immediately
%
% to
%         JSN   anywhere        inverse jump
%
% LapoptUnroll takes a sequence of LAP instructions and reforms ist.
% Jumps backwards are determined and the code juped to is copied to
% the jump point until a copy break is determied. Breaks are
%   the end of code
%       in this case the whole preceeding code has already been
%       copied (comple on level unroll),
%   a jump instrcution
%       this is copied to the end of the inserted part
%   a returnjump (function call)
%       this is copied as a "divided" instruction: the return address
%       is set to the return place copied from (insertion of a label
%       behind the R instruction) and the call takes place by a jump.
%
% The second parameter tells, how often a completely unrollable loop
% is allowed to be unrolled.
%
% Makes sense preceeding to all other code optimization by enlarging
% the basic blocks.
%
% Especially useful with explicit and implicit loops (e.g. tail
% recursion).
%
% For Cray instruction set only
%
 
(compiletime (progn (load useful) (on fast-integers)))
 
(fluid '(*ConditionalJumps*))
 
(compiletime
  (ds addInstr ($i)
     (let (($x (cons $i nil)))
          (rplacd rcodeptr $x)
          (setq rcodeptr $x) ))
 
  (defmacro loop body `(while t . ,body))  )
(setq *conditionalJumps*
                              '((JAZ . JAN) (JAN . JAZ)
                                (JAP . JAM) (JAM . JAP)
                                (JSZ . JSN) (JSN . JSZ)
                                (JSP . JSM) (JSM . JSP)))
 
(de LapOptReformJumps (code)
  (let ((rcode code) x y z cond jump inv lab)
     (loop (when (null code) (return rcode))
          %              look for the reform condition
          (cond ((and
                     (pairp (setq cond (car code)))
                     (pairp (setq x (cdr code)))
                     (pairp (setq jump (car x)))
                     (eq (car jump) 'J)
                     (not (eqcar (cadr jump) 'Reg)) %exclude Reg jumps
                     (pairp (setq y (cdr x)))
                     (atom (setq lab (car y)))
                     (eq lab (cadr cond))
                     (setq inv
                          (atsoc (car cond) *ConditionalJumps*)) )
                (rplaca code (cons (cdr inv) (cdr jump)))
                (rplacd code (cddr code))  )
               ((and
                     (pairp (setq cond (car code)))
                     (atom (setq lab (cadr cond)))
                     (setq lab (memq lab code))
                     (setq jump (cadr lab))
                     (eqcar jump 'J)
                     (not (eq (caadr jump) 'Reg))
                (rplaca code (cons (car cond) (cdr jump)))
                )))
          (setq code (cdr code))
)))
 
(de LapOptUnRoll (code level)
   (prog (rcode rcodeptr instr instr2 lbl ptr continue limit nxtlbl x y
          unrollcount startofunroll      rptr)
      (when (not (and (> level -1) (< level 11)))
            (setq level 1))
      (setq rcode (setq rcodeptr (cons nil nil)))   % anchor for return val
 
      (loop (when (null code)(return nil))
            (setq instr (car code))
 
      (if (not (and (pairp instr)
                    (eq (car instr) 'J)
                    (atom (setq lbl (cadr instr)))
                    (or
                       (setq rptr (memq lbl rcode))
                       (setq ptr (memq lbl code))) ))
       (progn
               (addinstr instr)
               (setq code (cdr code)) )
 
          (if rptr
            (progn
               (setq unrollcount level)
               (setq nxtlbl (cadr code))    % next label
               (setq rptr (setq startofunroll (cdr rptr)))
               (setq continue t)
               (setq limit nil)
               (loop
                  (when (not continue) (return nil))
                  (when (eq rptr limit)          % start next cycle
                        (setq rptr startofunroll)
                        (setq unrollcount (- unrollcount 1))
                        (when (< unrollcount 0) (return nil)))
                  (setq instr2 (car rptr))
                  (cond
                    ((atom instr2)(setq instr2 nil)    % do not copy label
                                  (addinstr (gensym)) )
                    ((or (eq (car instr2) 'R) (eq (car instr2) 'J))
                          (setq continue nil))
 
                             % conditional jump to following label
                    ((and
                          (eq unrollcount 0)
                          (eq (cadr instr2) nxtlbl)
                          (setq x
                            (atsoc (car instr2) *ConditionalJumps*)))
                             % invert contional jump
                             %   insert label
                          (rplacd rptr (cons (setq y (gensym)) (cdr rptr)))
                          (setq instr2 (list (cdr x) y))
                          (setq continue nil))
 
                    (t (addinstr instr2)
                       (setq instr2 nil)))
                  (when (not limit) (setq limit rcodeptr))
                  (setq rptr (cdr rptr)) )         % end copy loop
 
                        % handling of breaking instruction
             (cond
                  ((null instr2)                   % code copied completely
                        (addinstr instr))          % copy original jump
 
                  ((eq (car instr2) 'J)
                        (when (not (eq (cadr instr2) (cadr code)))
                              (addinstr instr2) ))
 
                  ((eq (car instr2) 'R)            % split return jump
                         (setq lbl (gensym))       %  new label
                         (rplacd rptr (cons (car rptr) (cdr rptr)))
                         (rplaca rptr lbl)
                         (addinstr `(ploc (Reg A6) ,lbl))
                         (addinstr '(Move (reg B00) (Reg A6)))
                         (addinstr (cons 'J (cdr instr2))) )
 
                                                   % instr prepared
                  (t
                        (addinstr instr2)) )       % copy original jump
 
              (setq code (cdr code))
        )                                  % progn
 
      % Handling forward  "Jumps to jumps"
 
          (progn
               (setq ptr (cdr ptr))    % intruction sequence jumped to
               (while (and ptr (atom ptr)) (setq ptr (cdr ptr)))
               (setq ptr (car ptr))    % the instruction
               (when (memq (car ptr)
                               '(J JAZ JAM JAN JAP JSN JSZ JSM JSP))
                     (setq instr ptr))
               (addinstr instr)
               (setq code (cdr code)))) % progn,if rptr
        )                                 % if not ...
       )                                   % outer loop
       (setq rcode (cdr rcode))
       (return rcode)
))
 
