%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 370-CMAC-A.SL - first half of patterns for IBM370 PSL cmacro expansion   %
%                                                                          %
% Author:      Jim McDonald                                                %
%              IMSSS                                                       %
%              Stanford University                                         %
% Date:        11 December 1982                                            %
%                                                                          %
% Based on HP-CMAC.SL of Mar 82 by Maguire, Griss, Lowder                  %
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson                       %
%                                                                          %
% Revision:    27 June 1982 -- split into two files (-A and -B).           %
%                                                                          %
% All functions have been moved out of this file.                          %
%       Only patterns remain.                                              %
%       See 370-CMAC-FNS.SL for functions used here.                       %
%       See 370-CMAC-B.SL   for remaining patterns.                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       CMACRO DEFINITIONS                                  %
%                                                                           %
% Remember that it IS legal to use other CMACROS in expansions              %
%  and to define "fake" cmacros, i.e. procedures, for common cases.         %
%                                                                           %
% If a new C-macro is not standard, e.g. (!*WFOO ArgOne ArgTwo),            %
%  then define its table and procedure.                                     %
%                                                                           %
% You may have to define some more "anyregs" or 'TerminalOperands           %
%  or define predicates to be CAREFUL.                                      %
%                                                                           %
% (DefCMacro !*Wfoo                                                         %
%      ...... )                                                             %
%                                                                           %
% (de !*Wfoo (Arg1 Arg2)                                                    %
%   (Expand2OperandCmacro Arg1 Arg2 '!*Wfoo))                               %
%                                                                           %
%  WARNING:  When writing CMACROS, note that Arg1 might be (Reg T1),        %
%            and Arg2 might be (Reg T2), due to Anyreg expansions.          %
%            Be sure not to clobber T1 or T2 if ArgOne or ArgTwo,           %
%            respectively, is used later.                                   %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               SHIFTS                                                      %
%                                                                           %
% AShift   -- arithmetic shift                                              %
%                                                                           %
%             Positive index shifts left and shifts in zeroes.              %
%             Negative index shifts right and sign extends.                 %
%             ArgOne <- ArgOne shifted by ArgTwo                            %
%                                                                           %
% WShift   -- logical shift                                                 %
%                                                                           %
%             Positive index shifts left  and shifts in zeroes.             %
%             Negative index shifts right and shifts in zeroes.             %
%             ArgOne <- ArgOne shifted by ArgTwo                            %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  Eight shift instructions are available, giving all combinations of       %
%  left or right, single or double, and arithmetic or logical.              %
%                                                                           %
%  The arithmetic shifts set the condition code.                            %
%  The arithmetic left shifts can cause fixed-point overflow.               %
%  The doubleword shifts cause a specification check if R1 is odd.          %
%                                                                           %
%  R1 is shifted by the number expressed in the low six bits of the second  %
%  operand address.                                                         %
%                                                                           %
%  = SLA  R1,D2(B2)         RS Shift Left  Single            CC [FPOV]      %
%  = SLL  R1,D2(B2)         RS Shift Left  Single Logical                   %
%  = SRA  R1,D2(B2)         RS Shift Right Single            CC             %
%  = SRL  R1,D2(B2)         RS Shift Right Single Logical                   %
%                                                                           %
%    SLDA R1,D2(B2)         RS Shift Left  Double            CC [FPOV,SPEC] %
%    SLDL R1,D2(B2)         RS Shift Left  Double Logical                   %
%    SRDA R1,D2(B2)         RS Shift Right Double            CC [SPEC]      %
%    SRDL R1,D2(B2)         RS Shift Right Double Logical                   %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*AShift             % Arithmetic shift.  +index = left shift
        %
        %  special cases
        %
        ((AnyP ZeroP))          % null shift
        %
        %  register by constant
        %    shifts greater than 31 are equivalent to shifting 31
        %    shifts less than -31 are equivalent to shifting -31
        %
        ((RegP    NegLeq31P)  (SRA ArgOne (Minus ArgTwo)))
        ((RegP       Leq31P)  (SLA ArgOne ArgTwo))
        ((RegP NegFullInumP)  (SRA ArgOne 31)   % max shift possible
                              (SameLineComment "Arithmetic shift by" ArgTwo))
        ((RegP PosFullInumP)  (SLA ArgOne 31)   % max shift possible
                              (SameLineComment "Arithmetic shift by" ArgTwo))
        %
        %  register by register
        %       We can't use a negative value in a register as a shift
        %       so test sign, then shift left or right by positive amount.
        %
        ((RegP RegP)
                (LCR (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                                % TempReg := complement of shift
                (BH TempLabel1) % jump if TempReg positive (shift negative)
                    % normal case, positive (left) shift...
                (SLA ArgOne (Indexed ArgTwo 0))
                (B  TempLabel2)
                    % reverse case, negative (right) shift...
                (!*Lbl (label TempLabel1))
                (SRA ArgOne (Indexed (RegNotUsedBy ArgOne ArgTwo) 0))
                (!*Lbl (label TempLabel2)))
        %
        %  register by memory
        %       A shift specified in memory is useless.
        %       Get it into a register (which is still a hassle).
        %       Safe to recur, since lower level sees all of ArgOne.
        %
        ((RegP   AnyP)
                (!*Move ArgTwo (RegNotUsedBy ArgOne))    % put shift into reg
                (!*AShift ArgOne (RegNotUsedBy ArgOne))) % and proceed...
        %
        %  memory by constant -- we can only shift results in registers
        %    Safe to recur, since (in this case!) lower level will not
        %      allocate temp regs.
        %
        ((AnyP FullInumP)  % FullInumP includes ZeroP, HalfInumP, etc.
                (!*Move ArgOne (RegNotUsedBy ArgOne))   % load safe reg
                (!*AShift (RegNotUsedBy ArgOne) ArgTwo) % shift in reg
                (!*Move (RegNotUsedBy ArgOne) ArgOne))  % put back
        %
        %  memory by memory -- need to get both quantity-to-be-shifted and
        %     number-of-bits-to-shift into safe registers.
        %    Note:  it is NOT safe to recur--the lower level would allocate
        %     temp regs that could destroy ArgOne.
        %
        (   (!*Move ArgTwo (RegNotUsedBy ArgOne))       % first  !!
            (!*Move ArgOne (SecondRegNotUsedBy ArgOne)) % second !!
                              % Note: the second move can destroy ArgTwo
            (LCR (ThirdRegNotUsedBy ArgOne) (RegNotUsedBy ArgOne))
                              % TempReg := complement of shift
            (BH TempLabel1)   % jump if reg is positive (i.e., shift is neg)
                % normal case, positive (left) shift...
            (SLA ArgOne (Indexed (RegNotUsedBy ArgOne) 0))
            (B  TempLabel2)
                % reverse case, negative (right) shift...
            (!*Lbl (label TempLabel1))
            (SRA ArgOne (Indexed (ThirdRegNotUsedBy ArgOne) 0))
            (!*Lbl (label TempLabel2))
            (!*Move (SecondRegNotUsedBy ArgOne) ArgOne)))

(DefCMacro !*WShift               % Logical shift. +index = left shift
        %
        %  special cases
        %
        ((AnyP ZeroP))
        %
        %  register by constant -- shifts exceeding 31 zero out register
        %
        ((RegP NegLeq31P)       (SRL ArgOne (Minus ArgTwo)))
        ((RegP    Leq31P)       (SLL ArgOne ArgTwo))
        ((RegP FullInumP)       (SR ArgOne ArgOne)
                                (SameLineComment "All bits shifted away"))
        %
        %  register by register
        %       We can't use a negative value in a register as a shift
        %       so test sign, then shift left or right by positive amount.
        %
        ((RegP RegP)
                (LCR (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                                % TempReg := complement of shift
                (BH TempLabel1) % jump if TempReg positive (shift negative)
                    % normal case, positive (left) shift...
                (SLL ArgOne (Indexed ArgTwo 0))
                (B  TempLabel2)
                    % reverse case, negative (right) shift...
                (!*Lbl (label TempLabel1))
                (SRL ArgOne (Indexed (RegNotUsedBy ArgOne ArgTwo) 0))
                (!*Lbl (label TempLabel2)))
        %
        %  register by memory
        %       A shift specified in memory is useless.
        %       Get it into a register (which is still a hassle).
        %       Safe to recur, since lower level sees all of ArgOne.
        %
        ((RegP   AnyP)
                (!*Move ArgTwo (RegNotUsedBy ArgOne))    % put shift into reg
                (!*WShift ArgOne (RegNotUsedBy ArgOne))) % and proceed...
        %
        %  memory by constant -- we can only shift results in registers
        %    Safe to recur, since (in this case!) lower level will not
        %      allocate temp regs.
        %  could check for shift  > 31 here and optimize
        ((AnyP FullInumP)  % FullInumP includes ZeroP, HalfInumP, etc.
                (!*Move ArgOne (RegNotUsedBy ArgOne))   % load safe reg
                (!*WShift (RegNotUsedBy ArgOne) ArgTwo) % shift in reg
                (!*Move (RegNotUsedBy ArgOne) ArgOne))  % put back
        %
        %  memory by memory -- need to get both quantity-to-be-shifted and
        %     number-of-bits-to-shift into safe registers.
        %    Note:  it is NOT safe to recur--the lower level would allocate
        %     temp regs that could destroy ArgOne.
        %
        (   (!*Move ArgTwo (RegNotUsedBy ArgOne))       % first  !!
            (!*Move ArgOne (SecondRegNotUsedBy ArgOne)) % second !!
                              % Note: the second move can destroy ArgTwo
            (LCR (ThirdRegNotUsedBy ArgOne) (RegNotUsedBy ArgOne))
                              % TempReg := complement of shift
            (BH TempLabel1)   % jump if reg is positive (i.e., shift is neg)
                % normal case, positive (left) shift...
            (SLL ArgOne (Indexed (RegNotUsedBy ArgOne) 0))
            (B  TempLabel2)
                % reverse case, negative (right) shift...
            (!*Lbl (label TempLabel1))
            (SRL ArgOne (Indexed (ThirdRegNotUsedBy ArgOne) 0))
            (!*Lbl (label TempLabel2))
            (!*Move (SecondRegNotUsedBy ArgOne) ArgOne)))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               BOOLEAN OPERATIONS                                          %
%                                                                           %
%  These all perform bitwise logical operations:                            %
%                                                                           %
%  WNot     -- Target := NOT Source                                         %
%  WAnd     -- Target := Target AND Source                                  %
%  WOr      -- Target := Target OR  Source                                  %
%  WXor     -- Target := Target EXCLUSIVEOR Source                          %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  All of these set the condition code:                                     %
%  Each is used in some cmacro.                                             %
%                                                                           %
%  = NR   R1,R2             RR And                                          %
%  = N    R1,D2(X2,B2)      RX And                                          %
%  = NC   D1(L,B1),D2(B2)   SS And (character)                              %
%  = NI   D1(B1),I2         SI And Immediate                                %
%                                                                           %
%  = OR   R1,R2             RR Or                                           %
%  = O    R1,D2(X2,B2)      RX Or                                           %
%  = OC   D1(L,B1),D2(B2)   SS Or (character)                               %
%  = OI   D1(B1),I2         SI Or Immediate                                 %
%                                                                           %
%  = XR   R1,R2             RR XOr                                          %
%  = X    R1,D2(X2,B2)      RX XOr                                          %
%  = XC   D1(L,B1),D2(B2)   SS XOr (character)                              %
%  = XI   D1(B1),I2         SI XOr Immediate                                %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(DefCMacro !*WNot       % (!*WNot dest source)

        %
        ((RegP FullInumP) (!*Move (sub1 (minus ArgTwo)) ArgOne))
        ((RegP      AnyP) (!*Move ArgTwo ArgOne)
                          (X ArgOne (lit (fullword -1))))
        (                 (!*Move ArgTwo (RegNotUsedBy ArgOne))
                          (X (RegNotUsedBy ArgOne) (lit (fullword -1)))
                          (!*Move (RegNotUsedBy ArgOne) ArgOne)))

(DefCMacro !*WAnd

        % special cases
        ((EqualP))
        ((RegP      FwAllOnesP))        % no-ops if all bits are on
        ((FullWordP FwAllOnesP))
        ((HalfWordP HwAllOnesP))
        ((ByteP     ByteAllOnesP))
        ((AnyP      ZeroP)        (!*Move 0 ArgOne))

        % register -- register
        ((RegP      RegP)         (NR ArgOne ArgTwo))

        % register -- immediate or memory
        ((RegP         FullInumP) (N ArgOne (Lit (Fullword ArgTwo))))
        ((RegP UnsignedFullInumP) (N ArgOne (Lit (Fullword (Hex ArgTwo)))))

        ((RegP  BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                    (RegNotUsedBy ArgOne ArgTwo))
                            (N  ArgOne
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgOne ArgTwo)
                                               0))
                            (SameLineComment ArgOne ":=" ArgOne "LAND" ArgTwo))
        ((RegP   FullWordP) (N ArgOne ArgTwo))
        ((RegP        AnyP) (!*Move ArgTwo (RegNotUsedBy ArgOne))
                            (NR ArgOne (RegNotUsedBy ArgOne)))

        % memory -- immediate
        ((ByteP     ByteInumP)  (NI ArgOne ArgTwo))
        ((HalfWordP HalfInumP)  (!*SS1 NC ArgOne (Lit (Halfword ArgTwo)) 2))
        ((HalfWordP UnsignedHalfInumP)
                                (!*SS1 NC ArgOne
                                      (Lit (Halfword (Hex ArgTwo))) 2))
        ((FullWordP FullInumP)  (!*SS1 NC ArgOne (Lit (Fullword ArgTwo)) 4))
        ((FullWordP UnsignedFullInumP)
                                (!*SS1 NC ArgOne
                                      (Lit (Fullword (Hex ArgTwo))) 4))

        % memory -- memory
        ((HalfWordP HalfWordP)  (!*SS1 NC ArgOne ArgTwo 2))
        ((FullWordP FullWordP)  (!*SS1 NC ArgOne ArgTwo 4))
        ((Mem256P   Mem256P)    (!*SS1 NC ArgOne ArgTwo nil))

        % memory -- register
        ((AnyP RegP)       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                           (NR (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                           (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne))

        % memory -- memory
        %  Safe to recur, since complete ArgOne is seen by lower level.
        %
        (       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                (!*WAnd ArgOne (RegNotUsedBy ArgOne))))

(DefCMacro !*WOr

        % special cases
        ((EqualP))
        ((AnyP      ZeroP))

        % register -- register
        ((RegP      RegP)       (OR ArgOne ArgTwo))

        % register -- immediate or memory
        ((RegP         FullInumP)  (O ArgOne (Lit (Fullword ArgTwo))))
        ((RegP UnsignedFullInumP)  (O ArgOne
                                      (Lit (Fullword (Hex ArgTwo)))))

        ((RegP  BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                    (RegNotUsedBy ArgOne ArgTwo))
                            (O  ArgOne
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgOne ArgTwo)
                                               0))
                            (SameLineComment ArgOne ":=" ArgOne "LOR" ArgTwo))
        ((RegP  FullWordP)  (O ArgOne ArgTwo))
        ((RegP       AnyP)  (!*Move ArgTwo (RegNotUsedBy ArgOne))
                            (OR ArgOne (RegNotUsedBy ArgOne)))

        % memory -- immediate
        ((ByteP     ByteInumP)  (OI ArgOne ArgTwo))
        ((HalfWordP HalfInumP)  (!*SS1 OC ArgOne (Lit (Halfword ArgTwo)) 2))
        ((HalfWordP UnsignedHalfInumP)
                                (!*SS1 OC ArgOne
                                       (Lit (Halfword (Hex ArgTwo))) 2))
        ((FullWordP FullInumP)  (!*SS1 OC ArgOne (Lit (Fullword ArgTwo)) 4))
        ((FullWordP UnsignedFullInumP)
                                (!*SS1 OC ArgOne
                                       (Lit (Fullword (Hex ArgTwo))) 4))

        % memory -- memory
        ((HalfWordP HalfWordP)  (!*SS1 OC ArgOne ArgTwo 2))
        ((FullWordP FullWordP)  (!*SS1 OC ArgOne ArgTwo 4))
        ((Mem256P   Mem256P)    (!*SS1 OC ArgOne ArgTwo nil))

        % memory -- register
        ((AnyP RegP)       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                           (OR (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                           (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne))

        % memory -- memory
        %  Safe to recur, since complete ArgOne is seen by lower level.
        %
        (       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                (!*WOr  ArgOne (RegNotUsedBy ArgOne))))

(DefCMacro !*WXOr

        % special cases
        ((AnyP      ZeroP))

        % register -- register
        ((RegP      RegP)       (XR ArgOne ArgTwo))

        % register -- immmdiate or memory
        ((RegP         FullInumP)  (X ArgOne (Lit (Fullword ArgTwo))))
        ((RegP UnsignedFullInumP)  (X ArgOne
                                      (Lit (Fullword (Hex ArgTwo)))))
        ((RegP  BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                    (RegNotUsedBy ArgOne ArgTwo))
                            (X  ArgOne
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgOne ArgTwo)
                                               0))
                            (SameLineComment ArgOne ":=" ArgOne "XOR" ArgTwo))
        ((RegP      FullWordP)  (X ArgOne ArgTwo))
        ((RegP      AnyP)       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                                (XR ArgOne (RegNotUsedBy ArgOne)))
        % memory -- immediate
        ((ByteP     ByteInumP)  (XI ArgOne ArgTwo))
        ((HalfWordP HalfInumP)  (!*SS1 XC ArgOne (Lit (Halfword ArgTwo)) 2))
        ((HalfWordP UnsignedHalfInumP)
                                (!*SS1 XC ArgOne
                                      (Lit (Fullword (Hex ArgTwo))) 2))
        ((FullWordP FullInumP)  (!*SS1 XC ArgOne (Lit (Fullword ArgTwo)) 4))
        ((FullWordP UnsignedFullInumP)
                                (!*SS1 XC ArgOne
                                      (Lit (Fullword (Hex ArgTwo))) 4))

        % memory -- memory
        ((HalfWordP HalfWordP)  (!*SS1 XC ArgOne ArgTwo 2))
        ((FullWordP FullWordP)  (!*SS1 XC ArgOne ArgTwo 4))
        ((Mem256P   Mem256P)    (!*SS1 XC ArgOne ArgTwo nil))

        % memory -- register
        ((AnyP RegP)       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                           (XR (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                           (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne))

        % memory -- memory
        %  Safe to recur, since complete ArgOne is seen by lower level.
        %
        (       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                (!*WXOr ArgOne (RegNotUsedBy ArgOne))))

(DefCmacro !*SS1)  % for building instructions with SS format, version 1
(DefCmacro !*SS2)  % for building instructions with SS format, version 2

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               ARITHMETIC OPERATIONS                                       %
%                                                                           %
% These all perform two's complement integer arithmetic operations:         %
%                                                                           %
% WAddressPlus2 -- Destination := Destination + Source (if Dest < 2**24)    %
% WPlus2        -- Destination := Destination + Source                      %
% WDifference   -- Destination := Destination - Source                      %
% WMinus        -- Destination := - Source                                  %
% WTimes2       -- Destination := Destination * Source                      %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  = AR   R1,R2             RR Add                           CC [FPOV]      %
%                                                                           %
%  = AR   R1,R2             RR Add                           CC [FPOV]      %
%  = A    R1,D2(X2,B2)      RX Add                           CC [FPOV]      %
%  = AH   R1,D2(X2,B2)      RX Add Halfword                  CC [FPOV]      %
%    ALR  R1,R2             RX Add Logical                   CC             %
%    AL   R1,D2(X2,B2)      RX Add Logical                   CC             %
%                                                                           %
%  = LA   R1,D2(X2,B2)      RX Load Address  (zeros out high order byte)    %
%                                                                           %
%  = SR   R1,R2             RR Subtract                      CC [FPOV]      %
%  = S    R1,D2(X2,B2)      RX Subtract                      CC [FPOV]      %
%  = SH   R1,D2(X2,B2)      RX Subtract Halfword             CC [FPOV]      %
%    SLR  R1,R2             RX Subtract Logical              CC             %
%    SL   R1,D2(X2,B2)      RX Subtract Logical              CC             %
%                                                                           %
%  = MR   R1,R2             RR Multiply                         [SPEC]      %
%  = M    R1,D2(X2,B2)      RX Multiply                         [SPEC]      %
%  = MH   R1,D2(X2,B2)      RX Multiply Halfword                            %
%                                                                           %
%    DR   R1,R2             RR Divide                           [SPEC,FPDV] %
%    D    R1,D2(X2,B2)      RX Divide                           [SPEC,FPDV] %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  !*WAddressPlus2 should be used only when the result is guaranteed
%   to be a positive integer less than 2**24.   LA zeros high order byte.
(DefCMacro !*WAddressPlus2   %  (!*WAddressPlus2 dest source)

        % special cases
        ((AnyP ZeroP))
        ((RegP DispInumP)       (LA ArgOne (Indexed ArgOne ArgTwo)))
        ((RegP AnyP)            (!*WPlus2 ArgOne ArgTwo))
        (       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                (!*WAddressPlus2 (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne)))

(DefCMacro !*WPlus2          %  (!*WPlus2 dest source)

        % special cases
        ((AnyP ZeroP))
        ((AnyP NegFullInumP)    (!*WDifference ArgOne (minus ArgTwo)))

        % register -- register
        ((RegP RegP)            (AR ArgOne ArgTwo))

        % register -- immediate
 % LA zeros out high byte of result register -- see !*WAddressPlus2
 %      ((RegP DispInumP)       (LA ArgOne (Indexed ArgOne ArgTwo)))
        ((RegP HalfInumP)       (AH ArgOne (Lit (Halfword ArgTwo))))
        ((RegP FullInumP)       (A  ArgOne (Lit (Fullword ArgTwo))))

        % register -- memory
        ((RegP HalfwordP)   (AH ArgOne ArgTwo))
        ((RegP BigOffsetP)  (!*Move (OffsetOf ArgTwo)
                                    (RegNotUsedBy ArgOne ArgTwo))
                            (A  ArgOne
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgOne ArgTwo)
                                               0))
                            (SameLineComment ArgOne ":=" ArgOne "+" ArgTwo))
        ((RegP FullwordP)   (A  ArgOne ArgTwo)
                            (SameLineComment ArgOne ":=" ArgOne "+" ArgTwo ))

        % ??
        ((RegP AnyP)            (PROBLEM "Wplus2" ArgOne ArgTwo))
        (       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                (!*WPlus2 (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne)))

(DefCMacro !*WDifference     %  (!*WDifference dest source)

        % special cases
        ((EqualP)               (!*Move 0 ArgOne))
        ((AnyP  ZeroP))
        ((AnyP  NegFullInumP)   (!*WPlus2 ArgOne (minus ArgTwo)))

        % register -- register
        ((RegP  RegP)           (SR ArgOne ArgTwo))

        % register -- immediate
        ((RegP  HalfInumP)      (SH ArgOne (Lit (Halfword ArgTwo))))
        ((RegP  FullInumP)      (S  ArgOne (Lit (Fullword ArgTwo))))

        ((RegP  HalfwordP)  (SH ArgOne ArgTwo))
        ((RegP BigOffsetP)  (!*Move (OffsetOf ArgTwo)
                                    (RegNotUsedBy ArgOne ArgTwo))
                            (S  ArgOne
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgOne ArgTwo)
                                               0))
                            (SameLineComment ArgOne ":=" ArgOne "-" ArgTwo))
        ((RegP  FullwordP)  (S  ArgOne ArgTwo)
                            (SameLineComment ArgOne ":=" ArgOne "-" ArgTwo ))

        % ??
        ((RegP  AnyP)           (PROBLEM "Difference" ArgOne ArgTwo))
        (       (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                (!*WDifference (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne)))

(DefCMacro !*WMinus         % (!*WMinus dest source)
        ((RegP      RegP)       (LCR ArgOne ArgTwo))
        ((RegP FullInumP)       (!*Move (minus ArgTwo) ArgOne))
        ((RegP      AnyP)       (!*Move ArgTwo ArgOne)
                                (LCR ArgOne ArgOne))
        (       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                (LCR (RegNotUsedBy ArgOne) (RegNotUsedBy ArgOne))
                (!*Move (RegNotUsedBy ArgOne) ArgOne)))

(DefCMacro !*WTimes2         %  (!*WTimes2 dest source)  16 bit(??)

        % special cases
        ((AnyP ZeroP)        (!*Move 0 ArgOne))
        ((AnyP OneP))
        ((AnyP MinusOneP)    (!*WMinus ArgOne ArgOne)
                             (SameLineComment ArgOne ":=" ArgOne "* -1"))
        ((AnyP PowerOf2P)    (!*AShift ArgOne (Log2 ArgTwo))
                         (SameLineComment ArgOne ":=" ArgOne "*" ArgTwo))

        % register -- register
        % first arg to MR must be the even register in an even-odd pair
        % Note that !*Move optimizes to a no-op if its args are equal
        ((RegT3P   RegP)     (MR (Reg T2) ArgTwo)
                             (SameLineComment "T2,T3 := T3 *" ArgTwo)
                             (SLDA (Reg T2) 32)
                             (SameLineComment "Interrupts if overflow")
                             (LR (Reg T3) (Reg T2))
                             (SameLineComment "T3 := product"))
        ((RegP   RegT1P)     (!*Move ArgOne (Reg T3))
                             (MR (Reg T2) ArgTwo)
                             (SameLineComment "T2,T3 := T3 *" ArgTwo)
                             (SLDA (Reg T2) 32)
                             (SameLineComment "Interrupts if overflow")
                             (!*Move (Reg T2) ArgOne)
                             (SameLineComment ArgOne ":= product"))
        ((RegP     RegP)     (!*Move ArgOne (Reg T1))
                             (MR (Reg T0) ArgTwo)
                             (SameLineComment "T0,T1 := T1 *" ArgTwo)
                             (SLDA (Reg T0) 32)
                             (SameLineComment "Interrupts if overflow")
                             (!*Move (Reg T0) ArgOne)
                             (SameLineComment ArgOne ":= product"))

        % register -- immediate
% avoid MH, since it precludes any reasonable test for overflow
%       ((RegP   HalfInumP)  (MH ArgOne (Lit (Halfword ArgTwo)))
%                        (SameLineComment ArgOne ":=" ArgOne "*" ArgTwo))
        ((RegT3P FullInumP)  (M  (Reg T2) (Lit (Fullword ArgTwo)))
                             (SameLineComment "T2,T3 := T3 *" ArgTwo)
                             (SLDA (Reg T2) 32)
                             (SameLineComment "Interrupts if overflow")
                             (LR (Reg T3) (Reg T2))
                             (SameLineComment "T3 := product"))
        ((RegP   FullInumP)  (!*Move ArgOne (Reg T1))
                             (M  (Reg T0) (Lit (Fullword ArgTwo)))
                             (SameLineComment "T0,T1 := T1 *" ArgTwo)
                             (SLDA (Reg T0) 32)
                             (SameLineComment "Interrupts if overflow")
                             (!*Move (Reg T0) ArgOne)
                             (SameLineComment ArgOne ":= product"))

        % register -- memory
        ((RegP   HalfwordP)  (MH ArgOne ArgTwo)
                          (SameLineComment ArgOne ":=" ArgOne "*" ArgTwo)
                          (SameLineComment "Warning: no overflow check"))

        ((RegT3P BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                     (RegNotUsedBy (Reg T3) ArgTwo))
                             (M (Reg T2)
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                           (RegNotUsedBy (Reg T3) ArgTwo)
                                               0))
                             (SameLineComment "T2,T3 := T3 *" ArgTwo)
                             (SLDA (Reg T2) 32)
                             (SameLineComment "Interrupts if overflow")
                             (LR (Reg T3) (Reg T2))
                             (SameLineComment "T3 := product"))
        ((RegT3P FullWordP)  (M  (Reg T2) ArgTwo)
                             (SameLineComment "T2,T3 := T3 *" ArgTwo)
                             (SLDA (Reg T2) 32)
                             (SameLineComment "Interrupts if overflow")
                             (LR (Reg T3) (Reg T2))
                             (SameLineComment "T3 := product"))
        ((RegP BigOffsetP)   (!*Move ArgOne (Reg T1))
                             (!*Move (OffsetOf ArgTwo)
                                     (RegNotUsedBy (Reg T1) ArgTwo))
                             (M (Reg T0)
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                           (RegNotUsedBy (Reg T1) ArgTwo)
                                               0))
                             (SameLineComment "T0,T1 := T1 *" ArgTwo)
                             (SLDA (Reg T0) 32)
                             (SameLineComment "Interrupts if overflow")
                             (!*Move (Reg T0) ArgOne)
                             (SameLineComment ArgOne ":= product"))
        ((RegP   FullwordP)  (!*Move ArgOne (Reg T1))
                             (M  (Reg T0) ArgTwo)
                             (SameLineComment "T0,T1 := T1 *" ArgTwo)
                             (SLDA (Reg T0) 32)
                             (SameLineComment "Interrupts if overflow")
                             (!*Move (Reg T0) ArgOne)
                             (SameLineComment ArgOne ":= product"))

        % ??
        ((RegP AnyP)         (PROBLEM "WTimes2" ArgOne ArgTwo))

        % slightly inefficient -- recursive !*WTimes2 could just leave
        % results in even register, if there was some way to tell it.
        ((AvoidsT2T3P AnyP)  (!*Move ArgOne (Reg T3))
                             (!*WTimes2 (Reg T3) ArgTwo)
                             (!*Move (Reg T3) ArgOne))
        ((AvoidsT0T1P AnyP)  (!*Move ArgOne (Reg T1))
                             (!*WTimes2 (Reg T1) ArgTwo)
                             (!*Move (Reg T1) ArgOne))
        (       (PROBLEM "WTimes2" ArgOne ArgTwo)))

%
%  WQuotient and WRemainder are not easily expressed as open code,
%  so they are implemented as hand-coded routines in IBM370 ASSEMBLE.
%
%  !*WQuotient and !*WRemainder could be CMACRO's, though, to take
%  advantage of the fact that the divisor can be in a memory location.
%  (Also, regs 1 and 2 are suboptimal locations for the arguments.)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            ALLOCATION/DEALLOCATION/FLUID BINDINGS/PUSH/POP                %
%                                                                           %
% Alloc    -- Allocates stack space on procedure entry.                     %
% DeAlloc  -- Deallocates stack space at procedure exit.                    %
% FreeRstr -- Restores fluid bindings at prog/lambda exit.                  %
% Push     -- Push one object onto control stack.                           %
% Pop      -- Pop one object from control stack.                            %
%                                                                           %
% See arithmetic section for available machine instructions.                %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*Alloc                % ArgOne is expressed in bytes
        (       (*WDifference (Reg CtlStkR) Argone)))

(DefCMacro !*Dealloc              % ArgOne is expressed in bytes
        (       (*WAddressPlus2 (Reg CtlStkR) Argone)))
                                  % !*WAddressPlus2 is safe for addresses

(DefCMacro !*FreeRstr
        (       (!*CALL FastUnbind)
                (Halfword ArgOne)))

(DefCMacro !*Push
        (       (*WDifference (Reg CtlStkR) 4)      % stack grows down
                (*Move ArgOne (Frame 1))))

(DefCMacro !*Pop
        (       (*Move (Frame 1) ArgOne)
                (*WAddressPlus2 (Reg CtlStkR) 4)))  % stack shrinks up
                                  % !*WAddressPlus2 is safe for addresses


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       FetchCallInfo/SetCallInfo/PopCallInfo/PushCallInfo                  %
%                                                                           %
% WARNING:  These cmacros assume that the function calling stack has not    %
%           been altered since the beginning of the current function call,  %
%           except that balanced PUSH/POP's may have intervened.            %
%                                                                           %
% These two do NOT alter the stack:                                         %
%   FetchCallInfo -- Fetches return address from stack.                %
%   SetCallInfo   -- Changes stack location holding return address.    %
%                                                                           %
% These two WILL alter the stack:                                           %
%   PopCallInfo  -- Pops stack items used to implement function call.       %
%   PushCallInfo -- Pushes stack iterms used to implement function call.    %
%                                                                           %
% These cmacros help remove dependencies on the way function calls have     %
%  been implemented for a given machine architecture.                       %
% Because a function call may push one or more words on the stack, the      %
%  return address may or may not be the top item on the stack.              %
% On most systems, a function call pushes only the return address.          %
% On the IBM 370, a function call pushes the return address plus the        %
%  caller's base address, in such a way that !*Pop yields the caller's %    %
%  base address.                                                            %
%                                                                           %
% PopCallInfo is guaranteed to pop the return address into its first        %
%  argument, and auxiliary items into successive arguments.                 %
% PushCallInfo is provided so the effect of PopCallInfo can be undone.      %
%                                                                           %
% Because Pass-1-Lap uses APPLY, which does not work for fexprs,            %
% PopCallInfo and PushCallInfo always take exactly two arguments.           %
% The first is used for the return address, the second is a list of other   %
%   items to be pushed or popped in some system-dependent manner.           %
% System-specific code then expands these into some fixed number of args    %
%  for the CmacroPatternTable given in the DefCMacro.                       %
%                                                                           %
% On the IBM 370, there will be two arguments, expanded as follows:         %
%                                                                           %
%                                             ArgOne           ArgTwo       %
%                                             ------           ------       %
% (!*PopCallInfo (reg i) nil)                 (reg i)            nil        %
% (!*PopCallInfo (reg i) ((reg j)))           (reg i)          (reg j)      %
% (!*PopCallInfo (reg i) ((reg j) (reg k)))            error                %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*GetCallInfo    % get return address, base address

        % If there is just one argument, make it receive return address,
        ((AnyP NullP)  % note that (NullP '(Immediate nil)) is true
           (!*Move (Memory (Reg CtlStkR) 4) ArgOne)
           (SameLineComment ArgOne ":= return address"))

        % If the two arguments are registers of the form J+1, J then LM
        %  be used to retrieve both at once.
        (TwoDescendingRegisters
           (LM ArgTwo ArgOne (Indexed (Reg CtlStkR) 0))
           (SameLineComment ArgTwo ":= caller's base address,"
                            ArgOne ":= return address"))

        % always try
        (  (!*Move (Memory (Reg CtlStkR) 0) ArgTwo)
           (SameLineComment ArgTwo ":= Caller's base address")
           (!*Move (Memory (Reg CtlStkR) 4) ArgOne)
           (SameLineComment ArgOne ":= return address") )

) % end of GetCallInfo


(DefCMacro !*SetCallInfo    % set return address, base address

        % If there is just one argument, use it to set return address
        ((AnyP NullP)  % note that (NullP '(Immediate nil)) is true
           (!*Move ArgOne (Memory (Reg CtlStkR) 4))
           (SameLineComment "return address :=" ArgOne ))

        % If the two arguments are registers of the form J+1, J then STM
        %  be used to set both at once.
        (TwoDescendingRegisters
           (STM ArgTwo ArgOne (Indexed (Reg CtlStkR) 0))
           (SameLineComment "return address :=" ArgOne
                            "caller's base address :=" ArgTwo))

        (  (!*Move ArgOne (Memory (Reg CtlStkR) 4))
           (SameLineComment "return address :=" ArgOne)
           (!*Move ArgTwo (Memory (Reg CtlStkR) 0))
           (SameLineComment "caller's base address :=" ArgTwo))

) % end of SetCallInfo

(DefCMacro !*PopCallInfo    % pop return address, base address

        (  (!*GetCallInfo ArgOne ArgTwo)
           (!*WAddressPlus2 (Reg CtlStkR) 8)
           (SameLineComment "Pop base addr, return address")
))

(DefCMacro !*PushCallInfo    % push return address, base address

        ((AnyP NullP)
           (PROBLEM "PushCallInfo needs two values" ArgOne ArgTwo))

        (  (!*WDifference (Reg CtlStkR) 8)
           (SameLineComment "Make room for two items")
           (!*SetCallInfo ArgOne ArgTwo)))




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                        LINK/CALL/EXIT                                     %
%                                                                           %
% Link        -- Link to function indicated by Source, returning here.      %
%                Sets NargReg, may set LinkReg, may set CallReg.            %
%                (The function being called might be a "foreign function"   %
%                 and thus use a non-PSL calling convention.)               %
%                                                                           %
% LinkE       -- Link to function indicated by Source, don't return here.   %
%                Sets NargReg, may set LinkReg, may set CallReg.            %
%                (The function being called might be a "foreign function"   %
%                 and thus use a non-PSL calling convention.)               %
%                                                                           %
% Call        -- Call function indicated by Source, returning here.         %
%                Sets CallReg, not NargNeg, not LinkReg.                    %
%                                                                           %
% JCall       -- Call function indicated by Source, don't return here.      %
%                (Used for calls made at the end of a routine.)             %
%                Sets CallReg, not NargNeg, not LinkReg.                    %
%                                                                           %
% Exit        -- Return to caller.                                          %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  These do not set the condition code or cause any exceptions,             %
%  except for jumps to illegal memory addresses:                            %
%                                                                           %
%  = BALR R1,R2             RR Branch and Link                              %
%    BAL  R1,D2(X2,B2)      RX Branch and Link                              %
%                                                                           %
%  The standard subroutine call will be BALR RETR,CALLREG                   %
%                                                                           %
%  Each routine starts by pushing BASER and RETR on the stack, then sets    %
%  BASER for its own use.  RETR must be saved since it holds the return     %
%  address, and BASER must be saved since it holds the base address for     %
%  the calling routine (hence must be restored when we return).             %
%  These are the instructions used:                                         %
%                                                                           %
%       SH    Stk15,Hw8               (4 bytes) make room                   %
%       STM   Bas13,Ret14,0(Stk15)    (4 bytes) save BaseR and RetR         %
%       LR    Bas13,Call9             (2 bytes) establish addressibility    %
%                                                                           %
%  If we want to do a JCALL, we should avoid our pop at the end, and also   %
%  jump past their entry sequence, as follows:                              %
%                                                                           %
%       Load Bas13 instead of Call9                                         %
%        (destroys ability to resume here, but this is an exit jump)        %
%       BAL   RetR,10(Bas13)                                                %
%        (jumps past the push sequence given above)                         %
%                                                                           %
%  CCI (which calls CompiledCallingInterpreted) and NULLDEFN (which calls   %
%   UndefinedFunction) will use RETR to look back here and decipher the     %
%   calling sequence to determine the ID invoked.                           %
%                                                                           %
%  Also see branch instructions in next section, in particular:             %
%                                                                           %
%  = BR   R1                RR Branch always  (= BCR 17,R1)                 %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*Link

        % If internal routine is calling itself, then BaseR already
        %  holds target address.
        ((InternalCurrentEntryP)
           (LR (Reg CallReg) (Reg BaseR))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Link" ArgOne))

        ((InternallyCallableP)
           (!*Move (Lit (FullWord (InternalEntry ArgOne))) (Reg CallReg))
                % use literal, since SymFnc entry will not exist
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Link" ArgOne))

        % calls through function cells more than 4096 from SymFncR
        ((BigOffsetIDP)
           (!*Move (OffsetOf (FnCell ArgOne)) (Reg LinkReg))
           (L (Reg CallReg) (DoubleIndexed (Reg SymFncR) (Reg LinkReg) 0))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Link" ArgOne))

        % normal calls (to a named routine)
        ((IDP)
           (!*Move (FnCell ArgOne) (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Link" ArgOne))

        % computed calls to (Memory (Reg x) SymFnc)
        ((IndirectViaSymFncP)
           (!*Move ArgOne (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Computed indirect link" ArgOne))

        % presume computed calls to (Memory (Reg x) (Wconst 0))
        (  (!*Move (immediate ArgOne) (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Computed link" ArgOne)))


(DefCMacro !*LinkE
        % If internal routine is calling itself, then BaseR already
        %  holds target address.
        ((InternalCurrentEntryP)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail link" ArgOne))

        ((InternallyCallableP)
           (!*Move (Lit (FullWord (InternalEntry ArgOne))) (Reg BaseR))
                % use literal, since SymFnc entry will not exist
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail link" ArgOne))

        % calls through function cells more than 4096 from SymFncR
        ((BigOffsetIDP)
           (!*Move (OffsetOf (FnCell ArgOne)) (Reg LinkReg))
           (L (Reg BaseR) (DoubleIndexed (Reg SymFncR) (Reg LinkReg) 0))
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail link" ArgOne))

        % normal jcall (to a named routine)
        ((IDP)
           (!*Move (FnCell ArgOne) (Reg BaseR))        % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail link" ArgOne))

        % computed call to (Memory (Reg x) SymFnc)
        ((IndirectViaSymFncP)
           (!*Move ArgOne (Reg BaseR))                 % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Computed indirect tail link" ArgOne))

        % presume computed call to (Memory (Reg x) (Wconst 0))
        (  (!*Move (immediate ArgOne) (Reg BaseR))     % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Computed tail link" ArgOne)))

(DefCMacro !*Call

        % If internal routine is calling itself, then BaseR already
        %  holds target address.
        ((InternalCurrentEntryP)
           (LR (Reg CallReg) (Reg BaseR))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Call" ArgOne))

        % other internal functions, move A(xxx) into callreg
        ((InternallyCallableP)
           (!*Move (Lit (FullWord (InternalEntry ArgOne))) (Reg CallReg))
                % use literal, since SymFnc entry will not exist
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Call" ArgOne))

        % call through function cell more than 4096 from SymFncR
        ((BigOffsetIDP)
           (!*Move (OffsetOf (FnCell ArgOne)) (Reg CallReg))
           (L (Reg CallReg) (DoubleIndexed (Reg SymFncR) (Reg CallReg) 0))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Call" ArgOne))

        % normal call (to a named routine)
        ((IDP)
           (!*Move (FnCell ArgOne) (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Call" ArgOne))

        % computed call via (Memory (reg x) SymFnc)
        ((IndirectViaSymFncP)
           (!*Move ArgOne (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Computed indirect call" ArgOne))

        % presume computed call via (Memory (reg x) (wconst 0))
        (  (!*Move (Immediate ArgOne) (Reg CallReg))
           (BALR (Reg RetR) (Reg CallReg))
           (SameLineComment "Computed call" ArgOne)))

(DefCMacro !*JCall
        % If internal routine is calling itself, then BaseR already
        %  holds target address.
        ((InternalCurrentEntryP)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail call" ArgOne))

        ((InternallyCallableP)
           (!*Move (Lit (FullWord (InternalEntry ArgOne))) (Reg BaseR))
                % use literal, since SymFnc entry will not exist
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail call" ArgOne))

        % calls through function cells more than 4096 from SymFncR
        ((BigOffsetIDP)
           (!*Move (OffsetOf (FnCell ArgOne)) (Reg CallReg))
           (L (Reg BaseR) (DoubleIndexed (Reg SymFncR) (Reg CallReg) 0))
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail call" ArgOne))

        % normal jcall (to a named routine)
        ((IDP)
           (!*Move (FnCell ArgOne) (Reg BaseR))        % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Tail call" ArgOne))

        % computed jcall via (Memory (reg x) SymFnc)
        ((IndirectViaSymFncP)
           (!*Move ArgOne (Reg BaseR))                 % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Computed indirect tail call" ArgOne))

        % preumed computed jcall via (Memory (reg x) (Wconst 0))
        (  (!*Move (immediate ArgOne) (Reg BaseR))     % new addressibility
                                                       % (abandon it here)
           (BAL (Reg RetR) (Indexed (Reg BaseR) 10))   % enter past pushes
           (SameLineComment "Computed tail call" ArgOne)))

(DefCMacro !*Exit         % ArgOne is already expressed in bytes
  ((LM (Reg BaseR) (Reg RetR) (Indexed (Reg CtlStkR) ArgOne))
                          % retrieve base-register and return-register
                          % they follow the allocated slots in memory
   (!*WAddressPlus2 (Reg CtlStkR) (plus2 Argone 8))
                          % release allocated slots and space for regs
                          % (each register occupied 4 bytes)
                          % !*WAddressPlus2 is safe for addresses
   (BR (Reg RetR))
   (SameLineComment "Return to (ancestral) caller")))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                     JUMPS/BRANCHES                                        %
%                                                                           %
% These instructions perform various jumps, which may be conditional        %
%   upon the status of the condition code or some register.                 %
%                                                                           %
% Note that an IF test on the 370 normally requires two instructions:       %
%   one to set the condition code, a second to conditionally branch.        %
%                                                                           %
% Also note that many instructions set the condition code, including        %
%   those that perform arithmetic, logical, and movement operations.        %
%   There are also a number of "pure" comparison instructions.              %
%                                                                           %
%   (A good compiler can avoid unnecessary tests by tracking the            %
%    status of the condition code register.  This is NOT done in any        %
%    comprehensive fashion by this cmacro assembler, and would be a         %
%    good candidate for optimization efforts.)                              %
%                                                                           %
%  WARNING:  The code that expands these cmacros rearranges the             %
%            order of arguments.                                            %
%            Thus, IN THE PATTERN, the order of args for JumpEq is          %
%             Arg1, Arg2, Label.                                            %
%            But IN THE CALL, the order of args is Label, Arg1, Arg2.       %
%                                                                           %
%  Jump           -- Unconditional jump to ArgOne                           %
%  JumpEq         -- Jump to ArgThree if ArgOne EQ ArgTwo                   %
%  JumpNeq        -- Jump to ArgThree if ArgOne NE ArgTwo                   %
%  JumpIf         -- Jump to ArgThree if                                    %
%                        ArgOne is related to ArgTwo by the test ArgFour    %
%                     or ArgTwo is related to ArgOne by the test ArgFive    %
%                                                                           %
%  JumpWGeq       -- Jump to ArgThree if ArgOne GE ArgTwo                   %
%  JumpWGreaterP  -- Jump to ArgThree if ArgOne GT ArgTwo                   %
%  JumpWLeq       -- Jump to ArgThree if ArgOne LE ArgTwo                   %
%  JumpWLessP     -- Jump to ArgThree if Argone LT ArgTwo                   %
%                                                                           %
%  JumpType       --                                                        %
%  JumpNotType    --                                                        %
%  JumpInType     --                                                        %
%  JumpNotInType  --                                                        %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  = BCR  R1,R2             RR Branch on Condition                          %
%  = BC   M1,D2(X2,B2)      RX Branch on Condition                          %
%                                                                           %
%       These jump if any masked bits of the (4-bit) condition code are on. %
%       Common tests have mnemonics that 370 assemblers will recognize:     %
%                                                                           %
%         B,   BR   -- branch always                                        %
%                                                                           %
%       After comparisons:                                                  %
%                                                                           %
%       = BE,  BER  -- branch if equal                                      %
%       = BNE, BNER -- branch if not equal                                  %
%       = BH,  BHR  -- branch if high                                       %
%       = BNH, BNHR -- branch if not high                                   %
%       = BL,  BLR  -- branch if low                                        %
%       = BNL, BNLR -- branch if not low                                    %
%                                                                           %
%       After loads and arithmetic instructions:                            %
%                                                                           %
%         BZ,  BZR  -- branch if zero                                       %
%         BNZ, BNZR -- branch if not zero                                   %
%         BP,  BHP  -- branch if positive                                   %
%         BNP, BNPR -- branch if not positive                               %
%         BN,  BNR  -- branch if negative                                   %
%         BNN, BNNR -- branch if not negative                               %
%                                                                           %
%  ---------------------------------------------                            %
%                                                                           %
%    BCTR R1,R2             RR Branch on Count                              %
%    BCT  R1,D2(X2,B2)      RX Branch on Count                              %
%                                                                           %
%         These decrement R1 by 1, then jump unless the result is 0.        %
%                                                                           %
%    BXH  R1,R3,D2(B2)      RS Branch on Index High                         %
%    BXLE R1,R3,D2(B2)      RS Branch on Index Low or Equal                 %
%                                                                           %
%         These add R3 to R1, then compare R1 to either R3 (if it is odd)    %
%         or the register after R3 (if it is even).                         %
%                                                                           %
%         BXH  jumps if R1 is larger.                                       %
%         BXLE jumps if R1 is less than or equal.                           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*Lbl (ArgOne))

(DefCMacro !*Jump
        ((Atom)            (B ArgOne))      % internal labels before compile
        ((TaggedLabel)     (B ArgOne))      % compiler generated labels
        ((ImmediateP)      (B (unimmediate ArgOne)))
        (                  (!*Comment "Jump is presumably a JCALL")
                           (!*JCALL ArgOne)))

(DefCmacro !*JumpEq
        ((Equal)          (B ArgThree))
        (                 (!*JumpIf ArgOne ArgTwo ArgThree BE BE)))

(DefCmacro !*JumpNotEq
        ((Equal))
        (                 (!*JumpIf ArgOne ArgTwo ArgThree BNE BNE)))

(DefCMacro !*JumpIf                  %( JumpIf A B Lbl Test ReverseTest )
        ((RegP AnyP)      (!*Compare ArgOne ArgTwo)
                          (ArgFour ArgThree))
        ((AnyP RegP)      (!*Compare ArgTwo ArgOne)
                          (ArgFive ArgThree))
        (                 (!*Compare ArgOne ArgTwo)
                          (ArgFour ArgThree)))
%
% In the following, the call to JumpEq, for example, shows the label
% being passed as the first argument.  This is correct.  The expansion
% of that macro will reorder the arguments to match the pattern seen
% above for JumpEq.  Likewise for the other jumps.
%
% I find this needlessly confusing, but perhaps there was some reason
% for it.  At any, beware.
%

(DefCMacro !*JumpInType
        (     (!*FIELD (RegNotUsedBy ArgOne ArgTwo) ArgOne 0 8)
              (!*JUMPEQ ArgThree (RegNotUsedBy ArgOne ArgTwo) 16#FF)
              (!*JUMPWGreaterP TempLabel (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
              (!*JUMPWGEQ ArgThree (RegNotUsedBy ArgOne ArgTwo) 0)
              (!*LBL (label TEMPLABEL))))

(DefCMacro !*JumpNotInType
        (     (!*FIELD (RegNotUsedBy ArgOne ArgTwo) ArgOne 0 8)
              (!*JUMPEQ TempLabel (RegNotUsedBy ArgOne ArgTwo) 16#FF)
              (!*JUMPWGreaterP ArgThree (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
              (!*JUMPWLessP ArgThree (RegNotUsedBy ArgOne ArgTwo) 0)
              (!*LBL (label TEMPLABEL))))

(DefCMacro !*JumpNotType
        (     (!*FIELD (RegNotUsedBy ArgOne ArgTwo) ArgOne 0 8)
              (!*JUMPNOTEQ ArgThree (RegNotUsedBy ArgOne ArgTwo) ArgTwo)))

(DefCMacro !*JumpType
        (     (!*FIELD (RegNotUsedBy ArgOne ArgTwo) ArgOne 0 8)
              (!*JUMPEQ ArgThree (RegNotUsedBy ArgOne ArgTwo) ArgTwo)))

(DefCmacro !*JumpWGeq)
(DefCmacro !*JumpWGreaterP)
(DefCmacro !*JumpWLeq)
(DefCmacro !*JumpWLessp)
